@@ -1,4 +1,4 @@
-"""Session creation and datasets registration
+"""Session creation and datasets registration.
 
 The RegistrationClient provides an high-level API for creating experimentation sessions on Alyx
 and registering associated datasets.
@@ -12,21 +12,23 @@ register_session - Create a session on Alyx from local path and register any ALF
 register_files - Register a list of files to their respective sessions on Alyx
 """
 import pathlib
-import uuid
+from uuid import UUID
 from pathlib import Path, PurePosixPath
 import datetime
 import logging
-from uuid import UUID
 import itertools
 from collections import defaultdict
 from fnmatch import fnmatch
+import shutil
 
 import requests.exceptions
 
 from iblutil.io import hashfile
+from iblutil.util import Bunch
 
 import one.alf.io as alfio
-from one.alf.files import session_path_parts, get_session_path
+from one.alf.files import session_path_parts, get_session_path, folder_parts, filename_parts
+from one.alf.spec import is_valid
 import one.alf.exceptions as alferr
 from one.api import ONE
 from one.util import ensure_list
@@ -35,6 +37,56 @@ from one.webclient import no_cache
 _logger = logging.getLogger(__name__)
 
 
+def get_dataset_type(filename, dtypes):
+    """Get the dataset type from a given filename.
+
+    A dataset type is matched one of two ways:
+
+     1. the filename matches the dataset type filename_pattern;
+     2. if filename_pattern is empty, the filename object.attribute matches the dataset type name.
+
+    Parameters
+    ----------
+    filename : str, pathlib.Path
+        The filename or filepath.
+    dtypes : iterable
+        An iterable of dataset type objects with the attributes ('name', 'filename_pattern').
+
+    Returns
+    -------
+    The matching dataset type object for filename.
+
+    Raises
+    ------
+    ValueError
+        filename doesn't match any of the dataset types
+        filename matches multiple dataset types
+    """
+    dataset_types = []
+    if isinstance(filename, str):
+        filename = PurePosixPath(filename)
+    for dt in dtypes:
+        if not dt.filename_pattern.strip():
+            # If the filename pattern is null, check whether the filename object.attribute matches
+            # the dataset type name.
+            if is_valid(filename.name):
+                obj_attr = '.'.join(filename_parts(filename.name)[1:3])
+            else:  # will match name against filename sans extension
+                obj_attr = filename.stem
+            if dt.name == obj_attr:
+                dataset_types.append(dt)
+        # Check whether pattern matches filename
+        elif fnmatch(filename.name.lower(), dt.filename_pattern.lower()):
+            dataset_types.append(dt)
+    n = len(dataset_types)
+    if n == 0:
+        raise ValueError(f'No dataset type found for filename "{filename}"')
+    elif n >= 2:
+        raise ValueError('Multiple matching dataset types found for filename '
+                         f'"{filename}": {", ".join(map(str, dataset_types))}')
+    return dataset_types[0]
+
+
 class RegistrationClient:
     """
     Object that keeps the ONE instance and provides method to create sessions and register data.
@@ -43,7 +95,7 @@ class RegistrationClient:
         self.one = one
         if not one:
             self.one = ONE(cache_rest=None)
-        self.dtypes = self.one.alyx.rest('dataset-types', 'list')
+        self.dtypes = list(map(Bunch, self.one.alyx.rest('dataset-types', 'list')))
         self.registration_patterns = [
             dt['filename_pattern'] for dt in self.dtypes if dt['filename_pattern']]
         self.file_extensions = [df['file_extension'] for df in
@@ -51,23 +103,23 @@ class RegistrationClient:
 
     def create_sessions(self, root_data_folder, glob_pattern='**/create_me.flag', dry=False):
         """
-        Create sessions looking recursively for flag files
+        Create sessions looking recursively for flag files.
 
         Parameters
         ----------
         root_data_folder : str, pathlib.Path
-            Folder to look for sessions
+            Folder to look for sessions.
         glob_pattern : str
-            Register valid sessions that contain this pattern
+            Register valid sessions that contain this pattern.
         dry : bool
-            If true returns list of sessions without creating them on Alyx
+            If true returns list of sessions without creating them on Alyx.
 
         Returns
         -------
         list of pathlib.Paths
-            Newly created session paths
+            Newly created session paths.
         list of dicts
-            Alyx session records
+            Alyx session records.
         """
         flag_files = list(Path(root_data_folder).glob(glob_pattern))
         records = []
@@ -82,7 +134,7 @@ class RegistrationClient:
         return [ff.parent for ff in flag_files], records
 
     def create_session(self, session_path, **kwargs) -> dict:
-        """Create a remote session on Alyx from a local session path, without registering files
+        """Create a remote session on Alyx from a local session path, without registering files.
 
         Parameters
         ----------
@@ -99,7 +151,7 @@ class RegistrationClient:
         return self.register_session(session_path, file_list=False, **kwargs)[0]
 
     def create_new_session(self, subject, session_root=None, date=None, register=True, **kwargs):
-        """Create a new local session folder and optionally create session record on Alyx
+        """Create a new local session folder and optionally create session record on Alyx.
 
         Parameters
         ----------
@@ -111,16 +163,16 @@ class RegistrationClient:
         date : datetime.datetime, datetime.date, str
             An optional date for the session.  If None the current time is used.
         register : bool
-            If true, create session record on Alyx database
+            If true, create session record on Alyx database.
         **kwargs
             Optional arguments for RegistrationClient.register_session.
 
         Returns
         -------
         pathlib.Path
-            New local session path
+            New local session path.
         uuid.UUID
-            The experiment UUID if register is True
+            The experiment UUID if register is True.
 
         Examples
         --------
@@ -155,19 +207,22 @@ class RegistrationClient:
         session_path : str, pathlib.Path
             The session path to search
 
-        Returns
+        Yields
         -------
-        generator
-            Iterable of file paths that match the dataset type patterns in Alyx
+        pathlib.Path
+            File paths that match the dataset type patterns in Alyx
         """
         session_path = Path(session_path)
-        types = (x['filename_pattern'] for x in self.dtypes if x['filename_pattern'])
-        dsets = itertools.chain.from_iterable(session_path.rglob(x) for x in types)
-        return (x for x in dsets if x.is_file() and
-                any(x.name.endswith(y) for y in self.file_extensions))
+        for p in session_path.rglob('*.*.*'):
+            if p.is_file() and any(p.name.endswith(ext) for ext in self.file_extensions):
+                try:
+                    get_dataset_type(p, self.dtypes)
+                    yield p
+                except ValueError as ex:
+                    _logger.debug('%s', ex.args[0])
 
     def assert_exists(self, member, endpoint):
-        """Raise an error if a given member doesn't exist on Alyx database
+        """Raise an error if a given member doesn't exist on Alyx database.
 
         Parameters
         ----------
@@ -190,20 +245,24 @@ class RegistrationClient:
             Member does not exist on Alyx
         requests.exceptions.HTTPError
             Failed to connect to Alyx database or endpoint not found
+
+        Returns
+        -------
+        dict, list of dict
+            The endpoint data if member exists.
         """
-        if isinstance(member, (str, uuid.UUID)):
+        if isinstance(member, (str, UUID)):
             try:
-                self.one.alyx.rest(endpoint, 'read', id=str(member), no_cache=True)
+                return self.one.alyx.rest(endpoint, 'read', id=str(member), no_cache=True)
             except requests.exceptions.HTTPError as ex:
                 if ex.response.status_code != 404:
                     raise ex
                 elif endpoint == 'subjects':
                     raise alferr.AlyxSubjectNotFound(member)
                 else:
-                    raise alferr.ALFError(f'Member "{member}" doesn\'t exist in Alyx')
+                    raise alferr.ALFError(f'Member "{member}" doesn\'t exist in {endpoint}')
         else:
-            for x in member:
-                self.assert_exists(x, endpoint)
+            return [self.assert_exists(x, endpoint) for x in member]
 
     @staticmethod
     def ensure_ISO8601(date) -> str:
@@ -228,7 +287,7 @@ class RegistrationClient:
 
     def register_session(self, ses_path, users=None, file_list=True, **kwargs):
         """
-        Register session in Alyx
+        Register session in Alyx.
 
         NB: If providing a lab or start_time kwarg, they must match the lab (if there is one)
         and date of the session path.
@@ -342,34 +401,55 @@ class RegistrationClient:
             session['data_dataset_session_related'] = ensure_list(recs)
         return session, recs
 
-    def register_files(self, file_list, versions=None, default=True, created_by=None,
-                       server_only=False, repository=None, dry=False, max_md5_size=None):
+    def register_files(self, file_list,
+                       versions=None, default=True, created_by=None, server_only=False,
+                       repository=None, exists=True, dry=False, max_md5_size=None):
         """
-        Registers a set of files belonging to a session only on the server
+        Registers a set of files belonging to a session only on the server.
 
         Parameters
         ----------
         file_list : list, str, pathlib.Path
-            A filepath (or list thereof) of ALF datasets to register to Alyx
+            A filepath (or list thereof) of ALF datasets to register to Alyx.
+        versions : str, list of str
+            Optional version tags.
+        default : bool
+            Whether to set as default revision (defaults to True).
         created_by : str
-            Name of Alyx user (defaults to whoever is logged in to ONE instance)
-        repository : str
-            Name of the repository in Alyx to register to
+            Name of Alyx user (defaults to whoever is logged in to ONE instance).
         server_only : bool
             Will only create file records in the 'online' repositories and skips local repositories
-        versions : list of str
-            Optional version tags
-        default : bool
-            Whether to set as default revision (defaults to True)
+        repository : str
+            Name of the repository in Alyx to register to.
+        exists : bool
+            Whether the files exist on the repository (defaults to True).
         dry : bool
-            When true returns POST data for registration endpoint without submitting the data
+            When true returns POST data for registration endpoint without submitting the data.
         max_md5_size : int
-            Maximum file in bytes to compute md5 sum (always compute if None)
+            Maximum file in bytes to compute md5 sum (always compute if None).
+        exists : bool
+            Whether files exist in the repository. May be set to False when registering files
+            before copying to the repository.
 
         Returns
         -------
         list of dicts, dict
-            A list of newly created Alyx dataset records or the registration data if dry
+            A list of newly created Alyx dataset records or the registration data if dry.
+
+        Notes
+        -----
+        - The registered files may be automatically moved to new revision folders if they are
+         protected on Alyx, therefore it's important to check the relative paths of the output.
+        - Protected datasets are not checked in dry mode.
+        - In most circumstances a new revision will be added automatically, however if this fails
+         a 403 HTTP status may be returned.
+
+        Raises
+        ------
+        requests.exceptions.HTTPError
+            Submitted data not valid (400 status code)
+            Server side database error (500 status code)
+            Revision protected (403 status code)
         """
         F = defaultdict(list)  # empty map whose keys will be session paths
         V = defaultdict(list)  # empty map for versions
@@ -420,22 +500,168 @@ class RegistrationClient:
                   'hashes': md5s,
                   'filesizes': file_sizes,
                   'name': repository,
+                  'exists': exists,
                   'server_only': server_only,
                   'default': default,
-                  'versions': V[session_path]}
+                  'versions': V[session_path],
+                  'check_protected': True
+                  }
 
             # Add optional field
             if details['lab']:
                 r_['labs'] = details['lab']
             # If dry, store POST data, otherwise store resulting file records
-            records.append(r_ if dry else self.one.alyx.post('/register-file', data=r_))
+            try:
+                records.append(r_ if dry else self.one.alyx.post('/register-file', data=r_))
+            except requests.exceptions.HTTPError as err:
+                # 403 response when datasets already registered and protected by tags
+                err_message = err.response.json()
+                if not (err_message.get('status_code') == 403 and
+                   err_message.get('error') == 'One or more datasets is protected'):
+                    raise err  # Some other error occurred; re-raise
+                response = err_message['details']
+                today_revision = datetime.datetime.today().strftime('%Y-%m-%d')
+                new_file_list = []
+
+                for fl, res in zip(files, response):
+                    (name, prot_info), = res.items()
+                    # Dataset has not yet been registered
+                    if not prot_info:
+                        new_file_list.append(fl)
+                        continue
+
+                    # Check to see if the file path already has a revision in it
+                    file_revision = folder_parts(rel_path / fl, as_dict=True)['revision']
+                    # Find existing protected revisions
+                    existing_revisions = [k for pr in prot_info for k, v in pr.items() if v]
+
+                    if file_revision:
+                        # If the revision explicitly defined by the user doesn't exist or
+                        # is not protected, register as is
+                        if file_revision not in existing_revisions:
+                            revision_path = fl.parent
+                        else:
+                            # Find the next sub-revision that isn't protected
+                            new_revision = self._next_revision(file_revision, existing_revisions)
+                            revision_path = fl.parent.parent.joinpath(f'#{new_revision}#')
+
+                        if revision_path != fl.parent:
+                            session_path.joinpath(revision_path).mkdir(exist_ok=True)
+                            _logger.info('Moving %s -> %s', fl, revision_path.joinpath(fl.name))
+                            shutil.move(session_path / fl, session_path / revision_path / fl.name)
+                        new_file_list.append(revision_path.joinpath(fl.name))
+                        continue
+
+                    # The file wasn't in a revision folder but is protected
+                    fl_path = fl.parent
+                    assert name == fl_path.joinpath(fl.name).as_posix()
+
+                    # Find info about the latest revision
+                    # N.B on django side prot_info is sorted by latest revisions first
+                    (latest_revision, protected), = prot_info[0].items()
+
+                    # If the latest revision is the original and it is unprotected
+                    # no need for revision e.g {'clusters.amp.npy': [{'': False}]}
+                    if latest_revision == '' and not protected:
+                        # Use original path
+                        revision_path = fl_path
+
+                    # If there already is a revision but it is unprotected,
+                    # move into this revision folder e.g
+                    # {'clusters.amp.npy':
+                    #   [{'2022-10-31': False}, {'2022-05-31': True}, {'': True}]}
+                    elif not protected:
+                        # Check that the latest_revision has the date naming convention we expect
+                        # i.e. 'YYYY-MM-DD'
+                        try:
+                            _ = datetime.datetime.strptime(latest_revision[:10], '%Y-%m-%d')
+                            revision_path = fl_path.joinpath(f'#{latest_revision}#')
+                        # If it doesn't it probably has been made manually so we don't want to
+                        # overwrite this and instead use today's date
+                        except ValueError:
+                            # NB: It's possible that today's date revision is also protected but is
+                            # not the most recent revision. In this case it's safer to let fail.
+                            revision_path = fl_path.joinpath(f'#{today_revision}#')
+
+                    # If protected and the latest protected revision is from today we need to make
+                    # a sub-revision
+                    elif protected and today_revision in latest_revision:
+                        if latest_revision == today_revision:  # iterate from appending 'a'
+                            new_revision = self._next_revision(today_revision, existing_revisions)
+                        else:  # assume the revision is date + character, e.g. '2020-01-01c'
+                            alpha = latest_revision[-1]  # iterate from this character
+                            new_revision = self._next_revision(
+                                today_revision, existing_revisions, alpha)
+                        revision_path = fl_path.joinpath(f'#{new_revision}#')
+
+                    # Otherwise cases move into revision from today
+                    # e.g {'clusters.amp.npy': [{'': True}]}
+                    # e.g {'clusters.amp.npy': [{'2022-10-31': True}, {'': True}]}
+                    else:
+                        revision_path = fl_path.joinpath(f'#{today_revision}#')
+
+                    # Only move for the cases where a revision folder has been made
+                    if revision_path != fl_path:
+                        session_path.joinpath(revision_path).mkdir(exist_ok=True)
+                        _logger.info('Moving %s -> %s', fl, revision_path.joinpath(fl.name))
+                        shutil.move(session_path / fl, session_path / revision_path / fl.name)
+                    new_file_list.append(revision_path.joinpath(fl.name))
+
+                assert len(new_file_list) == len(files)
+                r_['filenames'] = [p.as_posix() for p in new_file_list]
+                r_['filesizes'] = [session_path.joinpath(p).stat().st_size for p in new_file_list]
+                r_['check_protected'] = False  # Speed things up by ignoring server-side checks
+
+                records.append(self.one.alyx.post('/register-file', data=r_))
+                files = new_file_list
+
             # Log file names
             _logger.info(f'ALYX REGISTERED DATA {"!DRY!" if dry else ""}: {rel_path}')
             for p in files:
-                _logger.info(f"ALYX REGISTERED DATA: {p}")
+                _logger.info(f'ALYX REGISTERED DATA: {p}')
 
         return records[0] if len(F.keys()) == 1 else records
 
+    @staticmethod
+    def _next_revision(revision: str, reserved: list = None, alpha: str = 'a') -> str:
+        """
+        Return the next logical revision that is not already in the provided list.
+        Revisions will increment by appending a letter to a date or other identifier.
+
+        Parameters
+        ----------
+        revision : str
+            The revision on which to base the new revision.
+        reserved : list of str
+            A list of reserved (i.e. already existing) revision strings.
+        alpha : str
+            The starting character as an integer, defaults to 'a'.
+
+        Returns
+        -------
+        str
+            The next logical revision string that's not in the reserved list.
+
+        Examples
+        --------
+        >>> RegistrationClient._next_revision('2020-01-01')
+        '2020-01-01a'
+        >>> RegistrationClient._next_revision('2020-01-01', ['2020-01-01a', '2020-01-01b'])
+        '2020-01-01c'
+        >>> RegistrationClient._next_revision('2020-01-01', ['2020-01-01a', '2020-01-01b'])
+        '2020-01-01c'
+        """
+        if len(alpha) != 1:
+            raise TypeError(
+                f'`alpha` must be a character; received a string of length {len(alpha)}'
+            )
+        i = ord(alpha)
+        new_revision = revision + chr(i)
+        while new_revision in (reserved or []):
+            i += 1
+            new_revision = revision + chr(i)
+        return new_revision
+
     def register_water_administration(self, subject, volume, **kwargs):
         """
         Register a water administration to Alyx for a given subject
@@ -476,7 +702,7 @@ class RegistrationClient:
         # Ensure subject exists
         self.assert_exists(subject, 'subjects')
         # Ensure user(s) exist
-        user = ensure_list(kwargs.pop('user', [])) or self.one.alyx.user
+        user = kwargs.pop('user', self.one.alyx.user)
         self.assert_exists(user, 'users')
         # Ensure volume not zero
         if volume == 0:
@@ -500,14 +726,14 @@ class RegistrationClient:
 
     def register_weight(self, subject, weight, date_time=None, user=None):
         """
-        Register a subject weight to Alyx
+        Register a subject weight to Alyx.
 
         Parameters
         ----------
         subject : str
-            A subject nickname that exists on Alyx
+            A subject nickname that exists on Alyx.
         weight : float
-            The subject weight in grams
+            The subject weight in grams.
         date_time : str, datetime.datetime, datetime.date
             The time of weighing.  If None, the current time is used.
         user : str
